/*
 * This file contains the user specified measurement functions for the different integration regions
 * for the uncorrelated emission colour structure.
 * The recipe to derive the measurement functions F from the definition of the observable can be found 
 * in the manual.
 */

#include <Auxiliaries.h>
#include <Input_Common.h>
#include <Input_Parameters.h>
#include <cmath>
#include <iostream>

#ifdef BOOSTEDSOURCE_H_
using namespace boost::multiprecision;
using namespace boost::math;
#endif /* BOOSTEDSOURCE_H_ */

/*   __  __                                                    _      __                  _   _
 *  |  \/  |                                                  | |    / _|                | | (_)
 *  | \  / | ___  __ _ ___ _   _ _ __ ___ _ __ ___   ___ _ __ | |_  | |_ _   _ _ __   ___| |_ _  ___  _ __  ___
 *  | |\/| |/ _ \/ _` / __| | | | '__/ _ \ '_ ` _ \ / _ \ '_ \| __| |  _| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  | |  | |  __/ (_| \__ \ |_| | | |  __/ | | | | |  __/ | | | |_  | | | |_| | | | | (__| |_| | (_) | | | \__ \
 *  |_|  |_|\___|\__,_|___/\__,_|_|  \___|_| |_| |_|\___|_| |_|\__| |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 */

/*  _    _                               _       _           _ 
 * | |  | |                             | |     | |         | |
 * | |  | |_ __   ___ ___  _ __ _ __ ___| | __ _| |_ ___  __| |
 * | |  | | '_ \ / __/ _ \| '__| '__/ _ \ |/ _` | __/ _ \/ _` |
 * | |__| | | | | (_| (_) | |  | | |  __/ | (_| | ||  __/ (_| |
 *  \____/|_| |_|\___\___/|_|  |_|  \___|_|\__,_|\__\___|\__,_|
 */


double GAa(decimal r,decimal y,decimal tkl,decimal tk,decimal tl,decimal b) {
	// Please use the predefined variables tkl (=t_kl), ck (=cos theta_k) or tk (=(1-ck)/2), cl (=cos theta_l) or tl (=(1-cl)/2) for the definitions below
	decimal ck=Ck(tk);
	decimal cl=Cl(tl);
	decimal GAa=0;

	/* Please input the C++ syntax measurement function G_Aa here. As a reminder: G_Aa you get by using the
	 * parametrisation for the uncorrelated emission from the manual to the exponent of the full 
	 * measurement exponential, resubstituting yk-> r*y, yl->y and discarding a few factors.
	 *
	 * Don't forget to capture all zeroes and integrable divergences via if-clauses that set GAa to a constant.
 	 */ 

	//auxiliary variables containing the contributions from k and l emissions
	//kcontrib is problematic at r and y ->0, lcontrib only at y->0	
	decimal kcontrib=4.*sb*(b*(sqrt(-4*(-1+pow(ck,2))*r*y+pow(-2*sqrt(1.-pow(sb,2))*ck*sqrt(r*y)+sb*(-1+r*y),2))-abs(sb-r*sb*y+2*sqrt(1.-pow(sb,2))*ck*sqrt(r*y))))/(2.*(1+b)*r*y*(1+y));
	decimal kcontribLimr=4.*b*(1.-pow(ck,2))*pow(1.+y,-1)*pow(1.+b,-1);
	decimal kcontribLimy=4.*b*(1.-pow(ck,2))*pow(1.+b,-1);
	decimal kcontribLimboth=4.*b*(1.-pow(ck,2))*pow(1.+b,-1);
	decimal lcontrib=4.*sb*(sqrt(pow(sb*(-1+y)-2*sqrt(1.-pow(sb,2))*cl*sqrt(y),2)-4*(-1+pow(cl,2))*y)-abs(sb+2*sqrt(1.-pow(sb,2))*cl*sqrt(y)-sb*y))/(2.*(1+b)*y*(1+r*y));
	decimal lcontribLimy=4.*(1.-pow(cl,2))*pow(1.+b,-1);

	if ( r<=0. && y<=0. ) {
		GAa=kcontribLimboth+lcontribLimy;
	} else if ( r<=0. ) {
		GAa=kcontribLimr+lcontrib;
	} else if ( y<=0. ) {
		GAa=kcontribLimy+lcontribLimy;
	} else {
		GAa=kcontrib+lcontrib;
	}

	if ( cl<=-1. || cl>=1. ) {
		GAa=1.;
	}


	// Automated error message if a zero/divergence is not defused, please leave this as it is
	if (GAa<=0. || isnan(GAa) || isinf(GAa)) {
		cerr<<"Measurement function zero or NaN in region A, subregion a with value "<<GAa<<" at y="<<y<<" r="<<r<<" b="<<b<<" tkl="<<tkl<<" tk="<<tk<<" tl="<<tl<<endl;
		GAa=1.;
	}

	return ReturnDouble(GAa);
}

double GAb(decimal r,decimal y,decimal tkl,decimal tk,decimal tl,decimal b) {
	// Please use the predefined variables tkl (=t_kl), ck (=cos theta_k) or tk (=(1-ck)/2), cl (=cos theta_l) or tl (=(1-cl)/2) for the definitions below
	decimal ck=Ck(tk);
	decimal cl=Cl(tl);
	decimal GAb=0;

	/* Please input the C++ syntax measurement function G_Ab here. As a reminder: G_Ab you get by using the
	 * parametrisation for the uncorrelated emission from the manual to the exponent of the full 
	 * measurement exponential, resubstituting yl-> r*y, yk->y (yk, yl swapped compared to GAa) and discarding a few factors.
	 *
	 * Don't forget to capture all zeroes and integrable divergences via if-clauses that set GAb to a constant.
 	 */ 
	
	//auxiliary variables containing the contributions from k and l emissions
	//lcontrib is problematic at r and y ->0, kcontrib only at y->0	
	decimal lcontrib=4.*sb*(sqrt(-4*(-1+pow(cl,2))*r*y+pow(-2*sqrt(1.-pow(sb,2))*cl*sqrt(r*y)+sb*(-1+r*y),2))-abs(sb-r*sb*y+2*sqrt(1.-pow(sb,2))*cl*sqrt(r*y)))/(2.*(1+b)*r*y*(1+y));
	decimal lcontribLimr=4.*(1.-pow(cl,2))*pow(1+b,-1)*pow(1+y,-1);
	decimal lcontribLimy=4.*(1.-pow(cl,2))*pow(1+b,-1);
	decimal lcontribLimboth=4.*(1.-pow(cl,2))*pow(1+b,-1);
	decimal kcontrib=4.*sb*(b*(sqrt(pow(sb*(-1+y)-2*sqrt(1.-pow(sb,2))*ck*sqrt(y),2)-4*(-1+pow(ck,2))*y)-abs(sb+2*sqrt(1.-pow(sb,2))*ck*sqrt(y)-sb*y)))/(2.*(1+b)*y*(1+r*y));
	decimal kcontribLimy=4.*b*(1.-pow(ck,2))*pow(1+b,-1);

	if ( r<=0. && y<=0. ) {
		GAb=lcontribLimboth+kcontribLimy;
	} else if ( r<=0. ) {
		GAb=lcontribLimr+kcontrib;
	} else if ( y<=0. ) {
		GAb=lcontribLimy+kcontribLimy;
	} else {
		GAb=lcontrib+kcontrib;
	}

	if ( cl<=-1. || cl>=1. ) {
		GAb=1.;
	}

	

	// Automated error message if a zero/divergence is not defused, please leave this as it is
	if (GAb<=0. || isnan(GAb) || isinf(GAb)) {
		cerr<<"Measurement function zero or NaN in region A, subregion b with value "<<GAb<<" at y="<<y<<" r="<<r<<" b="<<b<<" tkl="<<tkl<<" tk="<<tk<<" tl="<<tl<<endl;
		GAb=1.;
	}

	return ReturnDouble(GAb);
}

double GB(decimal yk,decimal yl,decimal tkl,decimal tk,decimal tl,decimal b) {
	// Please use the predefined variables tkl (=t_kl), ck (=cos theta_k) or tk (=(1-ck)/2), cl (=cos theta_l) or tl (=(1-cl)/2) for the definitions below
	decimal ck=Ck(tk);
	decimal cl=Cl(tl);
	decimal GB=0;

	/* Please input the C++ syntax measurement function G_B here. As a reminder: G_B you get by using the
	 * parametrisation for the uncorrelated emission from the manual to the exponent of the full 
	 * measurement exponential, inverting ONE of EITHER yk or yl, and discarding a few factors. 
	 *
	 * Don't forget to capture all zeroes and integrable divergences via if-clauses that set GB to a constant.
 	 */ 

	//auxiliary variables containing the contributions from k and l emissions
	//kcontrib is problematic at yk->0, lcontrib at yl->0	
	decimal kcontrib=4.*sb*(b*(-sqrt(pow(sb*(-1+yk)-2*sqrt(1.-pow(sb,2))*ck*sqrt(yk),2))+sqrt(pow(sb*(-1+yk)-2*sqrt(1.-pow(sb,2))*ck*sqrt(yk),2)-4*(-1+pow(ck,2))*yk)))/(2.*(1+b)*yk*(1+yl));
	decimal kcontribLim=4.*b*(1.-pow(ck,2))*pow(1+yl,-1)*pow(1+b,-1);
	decimal lcontrib=4.*sb*(-sqrt(pow(sb*(-1+yl)-2*sqrt(1.-pow(sb,2))*cl*sqrt(yl),2))+sqrt(pow(sb*(-1+yl)-2*sqrt(1.-pow(sb,2))*cl*sqrt(yl),2)-4*(-1+pow(cl,2))*yl))/(2.*(1+b)*(1+yk)*yl);
	decimal lcontribLim=4.*(1.-pow(cl,2))*pow(1+yk,-1)*pow(1+b,-1);

	if ( yk<=0. && yl<=0. ) {
		GB=kcontribLim+lcontribLim;
	} else if ( yk<=0. ) {
		GB=kcontribLim+lcontrib;
	} else if ( yl<=0. ) {
		GB=kcontrib+lcontribLim;
	} else {
		GB=kcontrib+lcontrib;
	}

	if ( cl<=-1. || cl>=1. ) {
		GB=1.;
	}


	// Automated error message if a zero/divergence is not defused, please leave this as it is
	if (GB<=0. || isnan(GB) || isinf(GB)) {
		cerr<<"Measurement function zero or NaN in region B with value "<<GB<<" at yk="<<yk<<" yl="<<yl<<" b="<<b<<" tkl="<<tkl<<" tk="<<tk<<" tl="<<tl<<endl;
		GB=1.;
	}

	return ReturnDouble(GB);
}

/*This file is part of SoftSERVE.
 *
 *   SoftSERVE is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   SoftSERVE is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with SoftSERVE.  If not, see <http://www.gnu.org/licenses/>.
 */
