/*
 * This file contains the difference structures arising from plus distributions tying together the Measurement function f, plus distribution denominators 
 * and contributing remainder/numerator functions for the integration at all regulator-orders of the 1-loop soft function. Unlike for 2-loop we haven't split the 
 * computation along the dimension of the integration domain, as at most 2-dimensional integrals appear. So this computations includes also terms which 
 * strictly only depend on one variable, integrated over two.
 * Example: IOLA1E1 ('I'ntegration for 'O'ne-'L'oop soft function, at 'A'lpha to the '1'st and 'E'ps to the '1'st order [M1 stands for 'M'inus '1'st]) contains various logs of the f  
 * function and other polynomials in such a way that the various logs provide enough regulator factors to arrive at eps^1 alpha^1. Accordingly, we see below that IOLAM1E0 
 * contains no logs, as here one +-distribution's delta part provides the one inverse power of alpha, and all other factors are expanded to leading order only. 
 * IOLA0E0 contains logs of f(0,tl). f(y,tl) does not appear, as without the delta-part of the y-plus-distribution no alpha-divergence 
 * would be there, and then only leading order expanded terms could appear here, i.e. no logs. IOLA1E0 is therefore the first term in which the full 
 * variable-dependence of f appears. Alternatively, expanding to higher eps orders can also introduce various log structures, see e.g. IOLA0E1.
 */

#include <Auxiliaries.h>
#include <cmath>
#include <Input_Common.h>

using namespace std;

double IOLA1E0(double x, double s) {
	return (1 - s)*s*(1 + s)*(1 + pow(s,2))*pow(x,-1)*(-(mexp*log(f(0,tl(s)))) + mexp*log(f(y(x),tl(s))) - mexp*log(1 + pow(x,mexp)) + x*pow(log(f(0,tl(s))),2))*pow(2 - pow(s,4),-0.5)*pow(2 - 2*pow(s,4) + pow(s,8),-0.5);
}

double IOLA1E1(double x, double s) {
	return -((1 - s)*s*(1 + s)*(1 + pow(s,2))*pow(x,-1)*(4*mexp*log(f(y(x),tl(s)))*log(1 + pow(x,mexp)) + 2*log(x)*log(f(0,tl(s)))*pow(mexp,2) - 2*log(x)*log(f(y(x),tl(s)))*pow(mexp,2) + 4*mexp*pow(log(f(0,tl(s))),2) + 
        (-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)))*
         (2*mexp*log(f(0,tl(s))) - 2*mexp*log(f(y(x),tl(s))) + 2*mexp*log(1 + pow(x,mexp)) - 2*x*pow(log(f(0,tl(s))),2)) - 4*x*pow(log(f(0,tl(s))),3) - 4*mexp*pow(log(f(y(x),tl(s))),2))*pow(2 - pow(s,4),-0.5)*
      pow(2 - 2*pow(s,4) + pow(s,8),-0.5))/2.;
}

double IOLA1E2(double x, double s) {
	return -((1 - s)*s*(1 + s)*(1 + pow(s,2))*pow(x,-1)*(2*log(x)*pow(mexp,2)*pow(log(f(0,tl(s))),2) + 4*mexp*pow(log(f(0,tl(s))),3) - 4*x*pow(log(f(0,tl(s))),4) + 4*mexp*log(1 + pow(x,mexp))*pow(log(f(y(x),tl(s))),2) - 
        2*log(x)*pow(mexp,2)*pow(log(f(y(x),tl(s))),2) + (-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)))*
         (4*mexp*log(f(y(x),tl(s)))*log(1 + pow(x,mexp)) + 2*log(x)*log(f(0,tl(s)))*pow(mexp,2) - 2*log(x)*log(f(y(x),tl(s)))*pow(mexp,2) + 4*mexp*pow(log(f(0,tl(s))),2) - 4*x*pow(log(f(0,tl(s))),3) - 4*mexp*pow(log(f(y(x),tl(s))),2)) - 
        4*mexp*pow(log(f(y(x),tl(s))),3) + ((2*mexp*log(f(0,tl(s))) - 2*mexp*log(f(y(x),tl(s))) + 2*mexp*log(1 + pow(x,mexp)) - 2*x*pow(log(f(0,tl(s))),2))*
           pow(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)),2))/2.)*pow(2 - pow(s,4),-0.5)*pow(2 - 2*pow(s,4) + pow(s,8),-0.5))/2.;
}

double IOLA0E0(double x, double s) {
	return 2*(1 - s)*s*(1 + s)*log(f(0,tl(s)))*(1 + pow(s,2))*pow(2 - pow(s,4),-0.5)*pow(2 - 2*pow(s,4) + pow(s,8),-0.5);
}

double IOLA0E1(double x, double s) {
	return 2*(1 - s)*s*(1 + s)*(1 + pow(s,2))*pow(x,-1)*(-(mexp*log(f(0,tl(s)))) + mexp*log(f(y(x),tl(s))) + 
     x*log(f(0,tl(s)))*(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8))) + 2*x*pow(log(f(0,tl(s))),2))*pow(2 - pow(s,4),-0.5)*
   pow(2 - 2*pow(s,4) + pow(s,8),-0.5);
}

double IOLA0E2(double x, double s) {
	return -((1 - s)*s*(1 + s)*(1 + pow(s,2))*pow(x,-1)*(2*mexp*pow(log(f(0,tl(s))),2) + (-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)))*
        (2*mexp*log(f(0,tl(s))) - 2*mexp*log(f(y(x),tl(s))) - 4*x*pow(log(f(0,tl(s))),2)) - 4*x*pow(log(f(0,tl(s))),3) - 2*mexp*pow(log(f(y(x),tl(s))),2) - 
       x*log(f(0,tl(s)))*pow(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)),2))*pow(2 - pow(s,4),-0.5)*pow(2 - 2*pow(s,4) + pow(s,8),-0.5));
}

double IOLA0E3(double x, double s) {
	return -((1 - s)*s*(1 + s)*(1 + pow(s,2))*pow(x,-1)*((-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)))*
        (2*mexp*pow(log(f(0,tl(s))),2) - 4*x*pow(log(f(0,tl(s))),3) - 2*mexp*pow(log(f(y(x),tl(s))),2)) + (4*(mexp*pow(log(f(0,tl(s))),3) - 2*x*pow(log(f(0,tl(s))),4) - mexp*pow(log(f(y(x),tl(s))),3)))/3. + 
       ((2*mexp*log(f(0,tl(s))) - 2*mexp*log(f(y(x),tl(s))) - 4*x*pow(log(f(0,tl(s))),2))*pow(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)),2))/2. - 
       (x*log(f(0,tl(s)))*pow(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)),3))/3.)*pow(2 - pow(s,4),-0.5)*pow(2 - 2*pow(s,4) + pow(s,8),-0.5));
}


double IOLAM1E0(double x, double s) {
	return 2*(1 - s)*s*(1 + s)*(1 + pow(s,2))*pow(2 - pow(s,4),-0.5)*pow(2 - 2*pow(s,4) + pow(s,8),-0.5);
}
double IOLAM1E1(double x, double s) {
	return 2*(1 - s)*s*(1 + s)*(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) + 2*log(f(0,tl(s))) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)))*(1 + pow(s,2))*pow(2 - pow(s,4),-0.5)*
   pow(2 - 2*pow(s,4) + pow(s,8),-0.5);
}
double IOLAM1E2(double x, double s) {
	return (1 - s)*s*(1 + s)*(1 + pow(s,2))*(4*log(f(0,tl(s)))*(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8))) + 4*pow(log(f(0,tl(s))),2) + 
     pow(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)),2))*pow(2 - pow(s,4),-0.5)*pow(2 - 2*pow(s,4) + pow(s,8),-0.5);
}
double IOLAM1E3(double x, double s) {
	return ((1 - s)*s*(1 + s)*(1 + pow(s,2))*(12*(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)))*pow(log(f(0,tl(s))),2) + 8*pow(log(f(0,tl(s))),3) + 
       6*log(f(0,tl(s)))*pow(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)),2) + 
       pow(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)),3))*pow(2 - pow(s,4),-0.5)*pow(2 - 2*pow(s,4) + pow(s,8),-0.5))/3.;
}
double IOLAM1E4(double x, double s) {
	return ((1 - s)*s*(1 + s)*(1 + pow(s,2))*(32*(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)))*pow(log(f(0,tl(s))),3) + 16*pow(log(f(0,tl(s))),4) + 
       24*pow(log(f(0,tl(s))),2)*pow(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)),2) + 
       8*log(f(0,tl(s)))*pow(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)),3) + 
       pow(-4*log(1 - s) - 4*log(s) - 4*log(1 + s) - 4*log(1 + pow(s,2)) - log(2 - pow(s,4)) - log(2 - 2*pow(s,4) + pow(s,8)),4))*pow(2 - pow(s,4),-0.5)*pow(2 - 2*pow(s,4) + pow(s,8),-0.5))/12.;
}
