/*
 * This file contains various auxiliary function definitions, as well as the structures collecting individuals terms 
 * integrated over a fixed number of variables into master formulae to be fed to the integrator.
 *
 * The definitions of the "physical parametrisation" variables (like b and y) in terms of "computing 
 * parametrisation" variables (like c and x) are listed here as well. This allows us to let the user specify the 
 * measurement function in the more compact and intuitive physical parametrisation. It also provides a convenient breaking 
 * point for the introduction of multi-precision variables in the case of the "boosted" branch. We know that any potentially  
 * problematic large cancellations occur inside the measurement function, and so it suffices to change any data types used  
 * for measurement function internal computations to multiprecision variables. Accordingly, we use the conversion functions  
 * to translate between "double" and the multi-precision "decimal" types:
 * The integrator provides a "double" type point in the "computing parametrisation"-parametrised integration space. The  
 * reparametrisation functions below convert these to a "decimal" type point in "physical parametrisation"-parametrised  
 * integration space (e.g. decimal b(double c) ). The measurement functions FXi accept these "decimal" type coordinates,  
 * perform internal computations using multi-precision operations, and return a "double" type result to be used in the full,  
 * "double"-valued, integrand function.
 *
 * Additionally, to spare the user the task of having to remember which definition of the angular variables is used and in  
 * which sector Xi of the measurement functions FXi, we define the cosini Ck and Cl of relevant angles betweek \vec{k_T} and  
 * \vec{l_T} and the one allowed distinguished transverse direction \vec{n_T} as functions of the relevant computing  
 * parametrisation variables.
 * 
 *
 * Finally, we define the integrand structures fed to the integrator. These are best described using an example:
 * For the 3-dimensional integration of the Cf^2 structure, there are 4 different expressions to be integrated over 3 variables,  
 * depending on which variable out  of (y,q,t5,b) or their rescaled versions is integrated over, in addition to the always 
 * present t/s and tl/s6.
 * Each of these expressions has an eps- and alpha-expansion, so I get a sum of 4 terms for each appearing (eps, alpha)-order.
 * 
 * Cuba now requires the integrand in a peculiar form: It requires an integer-valued function with two arrays as arguments:  
 * x[nvar] and F[ncomp] (where nvar is the number of integration variables, and ncomp is the number of eps-orders, i.e.  
 * number of different sums of four terms above), and fills the array F with the values of the integrand structures, evaluated  
 * at the point with coordinates x. If this succeeds, the integer it returns is 1. A few auxiliary arguments are included for  
 * these functions, but these are irrelevant for our purposes.
 *
 * The notation is such that "D3Int" is the complete 'D'imension '3' 'Int'egrand, and contains all 
 * regulator orders, ordered from least divergent in the first array entry to the most divergent in the last array entry, grouped
 * in powers of alpha. So the least divergent alpha-powers (i.e. alpha^0) form th first group of five terms (eps^0 in F[0] to eps^-4
 * in F[4]), followed by the alpha^-1 terms (alpha^-1 eps^1 in F[5] to alpha^-1 eps^-3 in F[9]), followed by the alpha^-2 terms 
 * (alpha^-2 eps^2 in F[10] to alpha^-2 eps^-2 in F[14]) 
 */

#include <Auxiliaries.h>
#include <cmath>
#include <string>
#include <2Int.h>
#include <2IntAM1.h>
#include <2IntAM2.h>
#include <3Int.h>
#include <3IntAM1.h>
#include <3IntAM2.h>
#include <4Int.h>
#include <4IntAM1.h>
#include <4IntAM2.h>
#include <5Int.h>
#include <5IntAM1.h>
#include <XFactor.h>
#include <Input_Common.h>

using namespace std;

double mexp=M(m);

decimal tl(double s6) {
	return 4*pow(s6,4) - 6*pow(s6,8) + 4*pow(s6,12) - pow(s6,16);
}

decimal q(double r) {
	return pow(r,mexp);
}

decimal y(double x) {
	return pow(x,mexp);
}

decimal t5(double s5) {
	return pow(s5,2);
}


decimal t(double s) {
	return 4*pow(s,4) - 6*pow(s,8) + 4*pow(s,12) - pow(s,16);
}

decimal b(double c) {
	return pow(c,2.);
}

decimal Tk1(decimal t, decimal t5, decimal tl) {
	decimal tk1;
	if (t>=1.) {
		tk1=1-tl;
	} else if (t<=0.) {
		tk1=tl;
	} else if (tl>=1.) {
		tk1=1-t;
	} else if (tl<=0.) {
		tk1=t;
	} else {
		tk1=t + tl - 2*t*tl - 2*(t5 - 1)*sqrt((1 - t)*t*(1 - tl)*tl);
	}
	return tk1;
}

decimal Tk2(decimal t, decimal t5, decimal tl) {
	decimal tk2;
	if (t>=1.) {
		tk2=1-tl;
	} else if (t<=0.) {
		tk2=tl;
	} else if (tl>=1.) {
		tk2=1-t;
	} else if (tl<=0.) {
		tk2=t;
	} else {
		tk2=t + tl - 2*t*tl - 2*(1 - t5)*sqrt((1 - t)*t*(1 - tl)*tl);
	}
	return tk2;
}

decimal Ck1(decimal t, decimal t5, decimal tl) {
	decimal ck1;
	if (t>=1.) {
		ck1=2.*tl-1.;
	} else if (t<=0.) {
		ck1=1.-2.*tl;
	} else if (tl>=1.) {
		ck1=2.*t-1.;
	} else if (tl<=0.) {
		ck1=1.-2.*t;
	} else {
		ck1=1.-2.*(t + tl - 2*t*tl + 2*(1 - t5)*sqrt((1 - t)*t*(1 - tl)*tl));
	}
	return ck1;
}

decimal Ck2(decimal t, decimal t5, decimal tl) {
	decimal ck2;
	if (t>=1.) {
		ck2=2.*tl-1.;
	} else if (t<=0.) {
		ck2=1.-2.*tl;
	} else if (tl>=1.) {
		ck2=2.*t-1.;
	} else if (tl<=0.) {
		ck2=1.-2.*t;
	} else {
		ck2=1.-2.*(t + tl - 2*t*tl - 2*(1 - t5)*sqrt((1 - t)*t*(1 - tl)*tl));
	}
	return ck2;
}

decimal Cl(decimal tl) {
	return 1.-2.*tl;
}

decimal Ck(decimal tk) {
	return 1.-2.*tk;
}

//get FXi from FX to hide structures from users, also swap q and r here (I used q for yk/yl, Guido used r, so internally I still use q, but G is defined using r)
double FA1a(decimal q,decimal y,decimal t,decimal t5,decimal tl,decimal b) {
	return ReturnDouble(1.+b)*GAa(q,y,t,Tk1(t,t5,tl),tl,b);
}

double FA1b(decimal q,decimal y,decimal t,decimal t5,decimal tl,decimal b) {
	return ReturnDouble(1.+b)*GAb(q,y,t,Tk1(t,t5,tl),tl,b);
}

double FA2a(decimal q,decimal y,decimal t,decimal t5,decimal tl,decimal b) {
	return ReturnDouble(1.+b)*GAa(q,y,t,Tk2(t,t5,tl),tl,b);
}

double FA2b(decimal q,decimal y,decimal t,decimal t5,decimal tl,decimal b) {
	return ReturnDouble(1.+b)*GAb(q,y,t,Tk2(t,t5,tl),tl,b);
}

//get exponent M from variable m to hide structures from users
double M(double m) {
	double mm=2/m;
	if (isnan(mm) || isinf(mm) || mm==0. || mm<=0.) {
		return 2.;
	} else {
		return mm;
	}	
}

int D2Int(const int* ndim, const double x[], const int* ncomp, double F[],void *userdata) {
	F[0]=I2E0A0(x[0],x[1]); //eps^0
	F[1]=I2EM1A0(x[0],x[1]); //eps^-1
	F[2]=I2EM2A0(x[0],x[1]); //eps^-2
	F[3]=I2EM3A0(x[0],x[1]); //eps^-3
	F[4]=I2EM4A0(x[0],x[1]); //eps^-4
	F[5]=I2E1AM1(x[0],x[1]); //eps^1 alpha^-1
	F[6]=I2E0AM1(x[0],x[1]); //eps^0 alpha^-1
	F[7]=I2EM1AM1(x[0],x[1]); //eps^-1 alpha^-1
	F[8]=I2EM2AM1(x[0],x[1]); //eps^-2 alpha^-1
	F[9]=I2EM3AM1(x[0],x[1]); //eps^-3 alpha^-1
	F[10]=I2E2AM2(x[0],x[1]); //eps^2 alpha^-2
	F[11]=I2E1AM2(x[0],x[1]); //eps^1 alpha^-2
	F[12]=I2E0AM2(x[0],x[1]); //eps^0 alpha^-2
	F[13]=I2EM1AM2(x[0],x[1]); //eps^-1 alpha^-2
	F[14]=I2EM2AM2(x[0],x[1]); //eps^-2 alpha^-2

	return 1;
}

int D3Int(const int* ndim, const double x[],const int* ncomp, double F[],void* userdata) {
	F[0]=I3E0A0r(x[0],x[1],x[2])+I3E0A0x(x[0],x[1],x[2])+I3E0A0c(x[0],x[1],x[2])+I3E0A0s(x[0],x[1],x[2]); //eps^0
	F[1]=I3EM1A0r(x[0],x[1],x[2])+I3EM1A0x(x[0],x[1],x[2])+I3EM1A0c(x[0],x[1],x[2])+I3EM1A0s(x[0],x[1],x[2]); //eps^-1
	F[2]=I3EM2A0r(x[0],x[1],x[2])+I3EM2A0x(x[0],x[1],x[2])+I3EM2A0c(x[0],x[1],x[2])+I3EM2A0s(x[0],x[1],x[2]); //eps^-2
	F[3]=I3EM3A0r(x[0],x[1],x[2])+I3EM3A0x(x[0],x[1],x[2])+I3EM3A0c(x[0],x[1],x[2])+I3EM3A0s(x[0],x[1],x[2]); //eps^-3
	F[4]=0; //eps^-4
	F[5]=I3E1AM1r(x[0],x[1],x[2])+I3E1AM1x(x[0],x[1],x[2])+I3E1AM1c(x[0],x[1],x[2])+I3E1AM1s(x[0],x[1],x[2]); //eps^1 alpha^-1
	F[6]=I3E0AM1r(x[0],x[1],x[2])+I3E0AM1x(x[0],x[1],x[2])+I3E0AM1c(x[0],x[1],x[2])+I3E0AM1s(x[0],x[1],x[2]); //eps^0 alpha^-1
	F[7]=I3EM1AM1r(x[0],x[1],x[2])+I3EM1AM1x(x[0],x[1],x[2])+I3EM1AM1c(x[0],x[1],x[2])+I3EM1AM1s(x[0],x[1],x[2]); //eps^-1 alpha^-1
	F[8]=I3EM2AM1r(x[0],x[1],x[2])+I3EM2AM1x(x[0],x[1],x[2])+I3EM2AM1c(x[0],x[1],x[2])+I3EM2AM1s(x[0],x[1],x[2]); //eps^-2 alpha^-1
	F[9]=0; //eps^-3 alpha^-1
	F[10]=I3E2AM2r(x[0],x[1],x[2])+I3E2AM2x(x[0],x[1],x[2])+I3E2AM2c(x[0],x[1],x[2])+I3E2AM2s(x[0],x[1],x[2]); //eps^2 alpha^-2
	F[11]=I3E1AM2r(x[0],x[1],x[2])+I3E1AM2x(x[0],x[1],x[2])+I3E1AM2c(x[0],x[1],x[2])+I3E1AM2s(x[0],x[1],x[2]); //eps^1 alpha^-2
	F[12]=I3E0AM2r(x[0],x[1],x[2])+I3E0AM2x(x[0],x[1],x[2])+I3E0AM2c(x[0],x[1],x[2])+I3E0AM2s(x[0],x[1],x[2]); //eps^0 alpha^-2
	F[13]=I3EM1AM2r(x[0],x[1],x[2])+I3EM1AM2x(x[0],x[1],x[2])+I3EM1AM2c(x[0],x[1],x[2])+I3EM1AM2s(x[0],x[1],x[2]); //eps^-1 alpha^-2
	F[14]=0; //eps^-2 alpha^-2

	return 1;
}


int D4Int(const int* ndim, const double x[],const int* ncomp, double F[],void* userdata) {
	F[0]=I4E0A0rx(x[0],x[1],x[2],x[3])+I4E0A0rc(x[0],x[1],x[2],x[3])+I4E0A0rs(x[0],x[1],x[2],x[3])+I4E0A0xc(x[0],x[1],x[2],x[3])+I4E0A0xs(x[0],x[1],x[2],x[3])+I4E0A0cs(x[0],x[1],x[2],x[3]); //eps^0
	F[1]=I4EM1A0rx(x[0],x[1],x[2],x[3])+I4EM1A0rc(x[0],x[1],x[2],x[3])+I4EM1A0rs(x[0],x[1],x[2],x[3])+I4EM1A0xc(x[0],x[1],x[2],x[3])+I4EM1A0xs(x[0],x[1],x[2],x[3])+I4EM1A0cs(x[0],x[1],x[2],x[3]); //eps^-1
	F[2]=I4EM2A0rx(x[0],x[1],x[2],x[3])+I4EM2A0rc(x[0],x[1],x[2],x[3])+I4EM2A0rs(x[0],x[1],x[2],x[3])+I4EM2A0xc(x[0],x[1],x[2],x[3])+I4EM2A0xs(x[0],x[1],x[2],x[3])+I4EM2A0cs(x[0],x[1],x[2],x[3]); //eps^-2
	F[3]=0; //eps^-3
	F[4]=0; //eps^-4
	F[5]=I4E1AM1rx(x[0],x[1],x[2],x[3])+I4E1AM1rc(x[0],x[1],x[2],x[3])+I4E1AM1rs(x[0],x[1],x[2],x[3])+I4E1AM1xc(x[0],x[1],x[2],x[3])+I4E1AM1xs(x[0],x[1],x[2],x[3])+I4E1AM1cs(x[0],x[1],x[2],x[3]); //eps^1 alpha^-1
	F[6]=I4E0AM1rx(x[0],x[1],x[2],x[3])+I4E0AM1rc(x[0],x[1],x[2],x[3])+I4E0AM1rs(x[0],x[1],x[2],x[3])+I4E0AM1xc(x[0],x[1],x[2],x[3])+I4E0AM1xs(x[0],x[1],x[2],x[3])+I4E0AM1cs(x[0],x[1],x[2],x[3]); //eps^0 alpha^-1
	F[7]=I4EM1AM1rx(x[0],x[1],x[2],x[3])+I4EM1AM1rc(x[0],x[1],x[2],x[3])+I4EM1AM1rs(x[0],x[1],x[2],x[3])+I4EM1AM1xc(x[0],x[1],x[2],x[3])+I4EM1AM1xs(x[0],x[1],x[2],x[3])+I4EM1AM1cs(x[0],x[1],x[2],x[3]); //eps^-1 alpha^-1
	F[8]=0; //eps^-2 alpha^-1
	F[9]=0; //eps^-3 alpha^-1
	F[10]=I4E2AM2rx(x[0],x[1],x[2],x[3])+I4E2AM2rc(x[0],x[1],x[2],x[3])+I4E2AM2rs(x[0],x[1],x[2],x[3])+I4E2AM2xc(x[0],x[1],x[2],x[3])+I4E2AM2xs(x[0],x[1],x[2],x[3])+I4E2AM2cs(x[0],x[1],x[2],x[3]); //eps^2 alpha^-2
	F[11]=I4E1AM2rx(x[0],x[1],x[2],x[3])+I4E1AM2rc(x[0],x[1],x[2],x[3])+I4E1AM2rs(x[0],x[1],x[2],x[3])+I4E1AM2xc(x[0],x[1],x[2],x[3])+I4E1AM2xs(x[0],x[1],x[2],x[3])+I4E1AM2cs(x[0],x[1],x[2],x[3]); //eps^1 alpha^-2
	F[12]=I4E0AM2rx(x[0],x[1],x[2],x[3])+I4E0AM2rc(x[0],x[1],x[2],x[3])+I4E0AM2rs(x[0],x[1],x[2],x[3])+I4E0AM2xc(x[0],x[1],x[2],x[3])+I4E0AM2xs(x[0],x[1],x[2],x[3])+I4E0AM2cs(x[0],x[1],x[2],x[3]); //eps^0 alpha^-2
	F[13]=0; //eps^-1 alpha^-2
	F[14]=0; //eps^-2 alpha^-2

	return 1;
}

int D5Int(const int* ndim, const double x[],const int* ncomp, double F[],void* userdata) {
	F[0]=I5E0A0r(x[0],x[1],x[2],x[3],x[4])+I5E0A0x(x[0],x[1],x[2],x[3],x[4])+I5E0A0c(x[0],x[1],x[2],x[3],x[4])+I5E0A0s(x[0],x[1],x[2],x[3],x[4]); //eps^0
	F[1]=I5EM1A0r(x[0],x[1],x[2],x[3],x[4])+I5EM1A0x(x[0],x[1],x[2],x[3],x[4])+I5EM1A0c(x[0],x[1],x[2],x[3],x[4])+I5EM1A0s(x[0],x[1],x[2],x[3],x[4]); //eps^-1
	F[2]=0; //eps^-2
	F[3]=0; //eps^-3
	F[4]=0; //eps^-4
	F[5]=I5E1AM1r(x[0],x[1],x[2],x[3],x[4])+I5E1AM1x(x[0],x[1],x[2],x[3],x[4])+I5E1AM1c(x[0],x[1],x[2],x[3],x[4])+I5E1AM1s(x[0],x[1],x[2],x[3],x[4]); //eps^1 alpha^-1
	F[6]=I5E0AM1r(x[0],x[1],x[2],x[3],x[4])+I5E0AM1x(x[0],x[1],x[2],x[3],x[4])+I5E0AM1c(x[0],x[1],x[2],x[3],x[4])+I5E0AM1s(x[0],x[1],x[2],x[3],x[4]); //eps^0 alpha^-1
	F[7]=0; //eps^-1 alpha^-1
	F[8]=0; //eps^-2 alpha^-1
	F[9]=0; //eps^-3 alpha^-1
	F[10]=0; //eps^2 alpha^-2
	F[11]=0; //eps^1 alpha^-2
	F[12]=0; //eps^0 alpha^-2
	F[13]=0; //eps^-1 alpha^-2
	F[14]=0; //eps^-2 alpha^-2

	return 1;
}
